using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Windows;
using System.Windows.Resources;
using System.Xml.XPath;

namespace WpfPresentation.DataModel
{
    /// <summary>
    /// Contains all of the topics and pages displayed in the application.
    /// </summary>
    public class TableOfContents : INotifyPropertyChanged
    {
        #region Data

        TocItem currentItem;
        bool isSettingCurrentItem;
        Uri source;
        ReadOnlyCollection<Topic> topics;

        #endregion // Data

        #region Constructor

        /// <summary>
        /// Initializes a new instance of TableOfContents.  Set the
        /// Source property to initialize the TOC with content.
        /// </summary>
        public TableOfContents()
        {
        }

        #endregion // Constructor

        #region Public Properties

        #region CurrentItem

        /// <summary>
        /// Gets/sets topic or topic page currently being viewed.
        /// </summary>
        public TocItem CurrentItem
        {
            get { return this.currentItem; }
            set
            {
                // Anti-recursion gate and redundancy check.
                if (this.isSettingCurrentItem || value == this.CurrentItem)
                    return;

                try
                {
                    this.isSettingCurrentItem = true;

                    TocItem oldCurrentItem = this.CurrentItem;
                    TocItem newCurrentItem = value;

                    if (oldCurrentItem != null)
                    {
                        oldCurrentItem.IsCurrentItem = false;
                        this.currentItem = null;
                    }

                    if (newCurrentItem != null)
                    {
                        newCurrentItem.IsCurrentItem = true;
                        this.currentItem = newCurrentItem;
                    }

                    this.OnPropertyChanged("CurrentItem");
                }
                finally
                {
                    this.isSettingCurrentItem = false;
                }
            }
        }

        #endregion // CurrentItem

        #region Source

        /// <summary>
        /// Gets/sets a Uri which points to the XML file containing the table of contents data.
        /// If this property is set more than once, an InvalidOperationException is thrown.
        /// </summary>
        public Uri Source
        {
            get { return this.source; }
            set
            {
                // In this app, the TableOfContent's source should only be set once.
                if (this.source != null)
                    throw new InvalidOperationException("Why is Source being set again?");

                if (value == null)
                    throw new ArgumentNullException("Null is not a valid value for Source.");

                this.source = value;

                this.Initialize();
            }
        }

        #endregion // Source

        #region Topics

        /// <summary>
        /// Returns all of the Topics in the TOC.
        /// </summary>
        public ReadOnlyCollection<Topic> Topics
        {
            get { return this.topics; }
        }

        #endregion // Topics

        #endregion // Public Properties

        #region Private Helpers

        #region Initialize

        /// <summary>
        /// Populates the collection of Topics and sets the first one as the "current item."
        /// </summary>
        void Initialize()
        {
            Debug.Assert(this.Source != null, "Source should be set.");

            // Load the table of contents into an XPathDocument.
            StreamResourceInfo info = Application.GetResourceStream(this.Source);
            XPathDocument tocDoc = new XPathDocument(info.Stream);
            XPathNavigator tocNavigator = tocDoc.CreateNavigator();

            // Create a Topic for every <Topic> element in the XML file.
            List<Topic> topicList = new List<Topic>();
            XPathNodeIterator topicIterator = tocNavigator.Select("/TableOfContents/Topic");
            foreach (XPathNavigator topicNavigator in topicIterator)
            {
                topicList.Add(new Topic(this, topicNavigator));
            }

            this.topics = new ReadOnlyCollection<Topic>(topicList);

            // Set the first topic as the "current item."
            if (this.topics.Count > 0)
                this.CurrentItem = this.topics[0];
        }

        #endregion // Initialize

        #endregion // Private Helpers

        #region INotifyPropertyChanged

        private void OnPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        public event PropertyChangedEventHandler PropertyChanged;

        #endregion // INotifyPropertyChanged
    }
}