using System;
using System.Diagnostics;
using System.Windows;
using System.Windows.Controls;

namespace WpfPresentation.DataModel
{
    /// <summary>
    /// An entity which exists in the Table of Contents, either
    /// a Topic or a TopicPage.  This class is abstract.
    /// </summary>
    public abstract class TocItem
    {
        #region Data

        UIElement displayContent;
        bool isInitialized;
        Uri pageUri;
        readonly TableOfContents tableOfContents;

        #endregion // Data

        #region Constructor

        protected TocItem(TableOfContents tableOfContents)
        {
            if (tableOfContents == null)
                throw new ArgumentNullException("tableOfContents");

            this.tableOfContents = tableOfContents;
        }

        #endregion // Constructor

        #region Public Properties

        #region DisplayContent

        /// <summary>
        /// Returns the visual elements displayed in the main content area of the UI
        /// to render this item.
        /// </summary>
        public UIElement DisplayContent
        {
            get
            {
                this.Verify();
                return this.displayContent;
            }
        }

        #endregion // DisplayContent

        #region IsCurrentItem

        /// <summary>
        /// Gets/sets whether this item is the "current item" in 
        /// the TableOfContents to which it belongs.
        /// </summary>
        public bool IsCurrentItem
        {
            get { return this.TableOfContents.CurrentItem == this; }
            set
            {
                if (this.IsCurrentItem != value)
                    this.TableOfContents.CurrentItem = value ? this : null;
            }
        }

        #endregion // IsCurrentItem

        #region PageUri

        /// <summary>
        /// Returns the Uri which points to the Page associated with this item.
        /// </summary>
        public Uri PageUri
        {
            get
            {
                this.Verify();
                return this.pageUri;
            }
        }

        #endregion // PageUri

        #region TableOfContents

        /// <summary>
        /// Returns the TableOfContents in which this item exists.
        /// </summary>
        public TableOfContents TableOfContents
        {
            get
            {
                this.Verify();
                return this.tableOfContents;
            }
        }

        #endregion // TableOfContents

        #region Title [abstract]

        /// <summary>
        /// Returns the display name of this item.
        /// </summary>
        public abstract string Title { get; }

        #endregion // Title [abstract]

        #endregion // Public Properties

        #region Protected Methods

        #region Initialize

        protected void Initialize(Uri pageUri)
        {
            this.pageUri = pageUri;

            // Load the item's Page and store a reference to its root visual element.
            // Then remove that element from the Page so that it can be parented to the
            // ContentControl in the main Window.  We do not display the entire Page because
            // a Page can only be hosted by a Window or Frame.  When binding a Frame's Content
            // property to a collection of Pages, it always displays the same Page (bug?).
            Page page = Application.LoadComponent(this.pageUri) as Page;
            this.displayContent = page.Content as UIElement;
            page.Content = null;

            this.isInitialized = true;
        }

        #endregion // Initialize

        #region Verify

        /// <summary>
        /// Throws an exception if this object is not initialized.
        /// </summary>
        [Conditional("DEBUG")]
        [DebuggerStepThrough]
        protected void Verify()
        {
            if (!this.isInitialized)
                throw new Exception("TocItem is not initialized.");
        }

        #endregion // Verify

        #endregion // Protected Methods
    }
}