using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Windows;
using System.Windows.Controls;
using System.Xml.XPath;

namespace WpfPresentation.DataModel
{
    /// <summary>
    /// Represents a topic reviewed in the presentation.  
    /// A Topic contains TopicPages, and is contained within a TableOfContents.
    /// </summary>
    public class Topic : TocItem
    {
        #region Data

        readonly string title;
        readonly ReadOnlyCollection<TopicPage> topicPages;

        #endregion // Data

        #region Constructor

        /// <summary>
        /// Initializes a new instance of Topic.
        /// </summary>
        /// <param name="tableOfContents">The TOC which contains this topic.</param>
        /// <param name="topicNavigator">An XPathNavigator which points to the topic info.</param>
        public Topic(TableOfContents tableOfContents, XPathNavigator topicNavigator)
            : base(tableOfContents)
        {
            if (topicNavigator == null)
                throw new ArgumentNullException("topicNavigator");

            this.title = topicNavigator.GetAttribute("Title", "");

            // Load this topic's "home page."
            string folder = topicNavigator.GetAttribute("Folder", "");
            string homePageFileName = topicNavigator.GetAttribute("HomePage", "");
            Uri homePageUri = CreatePageUri(folder, homePageFileName);
            base.Initialize(homePageUri);

            // Create the content pages owned by this topic.
            List<TopicPage> topicPages = this.CreateTopicPages(folder, topicNavigator);
            this.topicPages = new ReadOnlyCollection<TopicPage>(topicPages);
        }

        #endregion // Constructor

        #region Public Properties

        #region Title

        /// <summary>
        /// Returns the display name of this Topic.
        /// </summary>
        public override string Title
        {
            get 
            {
                base.Verify();
                return this.title; 
            }
        }

        #endregion // Title

        #region TopicPages

        /// <summary>
        /// Returns the TopicPages owned by this Topic.
        /// </summary>
        public ReadOnlyCollection<TopicPage> TopicPages
        {
            get 
            {
                base.Verify();
                return this.topicPages; 
            }
        }

        #endregion // TopicPages

        #endregion // Public Properties

        #region Private Helpers

        #region CreatePageUri

        static Uri CreatePageUri(string folder, string pageFile)
        {
            string path = String.Format("/Topics/{0}/{1}", folder, pageFile);
            return new Uri(path, UriKind.Relative);
        }

        #endregion // CreatePageUri

        #region CreateTopicPages

        List<TopicPage> CreateTopicPages(string folder, XPathNavigator topicNavigator)
        {
            List<TopicPage> topicPageList = new List<TopicPage>();
            XPathNodeIterator pageIterator = topicNavigator.Select("./Pages/Page");
            foreach (XPathNavigator pageNavigator in pageIterator)
            {
                string fileName = pageNavigator.GetAttribute("FileName", "");
                Uri uri = CreatePageUri(folder, fileName);
                TopicPage topicPage = new TopicPage(this, uri);
                topicPageList.Add(topicPage);
            }
            return topicPageList;
        }

        #endregion // CreateTopicPages

        #endregion // Private Helpers
    }
}